/*

Copyright (c) 2005 Christian Pellegrin <chripell@gmail.com>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.

Based on:

----

    vfsdemo.c -- synchronous operations

    (from Gnome VFS)

----

    FUSE: Filesystem in Userspace
    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>

    This program can be distributed under the terms of the GNU GPL.
    See the file COPYING.

----

*/

#ifdef linux
/* For pread()/pwrite() */
#define _XOPEN_SOURCE 500
#define _GNU_SOURCE 1
#endif

#define FUSE_USE_VERSION 22

#include <fuse.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <dirent.h>
#include <errno.h>
#include <sys/statfs.h>
#include <sys/types.h>
#include <sys/stat.h>


#include <string.h>
//#include <libgnome/libgnome.h>
#include <libgnomevfs/gnome-vfs.h>


/* globals */

char * mount_name;		/* the base of the mounted hierarcy */

/* print a VFS error code, its description, and the problem URI */
int print_error(GnomeVFSResult code, const gchar *uri)
{
  const gchar *error_desc;

  error_desc = gnome_vfs_result_to_string(code);
  g_printerr("error %d when accessing %s: %s\n", code, uri, error_desc);

  return (code);
}

/* converts gnome vfs error to fuse one */
int gvfs2fuse(GnomeVFSResult result) {
  int r;

  //g_debug("error: %s", gnome_vfs_result_to_string(result));
  
  switch(result) {
  case GNOME_VFS_OK:
    r = 0;
    break;
  case GNOME_VFS_ERROR_NOT_FOUND:
    r = ENOENT;
    break;
  case GNOME_VFS_ERROR_GENERIC:
    r = ESTALE;
    break;
  case GNOME_VFS_ERROR_INTERNAL:
    r = ESTALE;
    break;
  case GNOME_VFS_ERROR_BAD_PARAMETERS:
    r = EFAULT;
    break;
  case GNOME_VFS_ERROR_NOT_SUPPORTED:
    r = ENOTSUP;
    break;
  case GNOME_VFS_ERROR_IO:
    r = EIO;
    break;
  case GNOME_VFS_ERROR_CORRUPTED_DATA:
    r = EIO;
    break;
  case GNOME_VFS_ERROR_WRONG_FORMAT:
    r = ENOEXEC;
    break;
  case GNOME_VFS_ERROR_BAD_FILE:
    r = EBADF;
    break;
  case GNOME_VFS_ERROR_TOO_BIG:
    r = EFBIG;
    break;
  case GNOME_VFS_ERROR_NO_SPACE:
    r = ENOSPC;
    break;
  case GNOME_VFS_ERROR_READ_ONLY:
    r = EROFS;
    break;
  case GNOME_VFS_ERROR_INVALID_URI:
    r = EFAULT;
    break;
  case GNOME_VFS_ERROR_NOT_OPEN:
    r = ESTALE;
    break;
  case GNOME_VFS_ERROR_INVALID_OPEN_MODE:
    r = EINVAL;
    break;
  case GNOME_VFS_ERROR_ACCESS_DENIED:
    r = EACCES;
    break;
  case GNOME_VFS_ERROR_TOO_MANY_OPEN_FILES:
    r = EMFILE;
    break;
  case GNOME_VFS_ERROR_EOF:
    r = EPIPE;
    break;
  case GNOME_VFS_ERROR_NOT_A_DIRECTORY:
    r = ENOTDIR;
    break;
  case GNOME_VFS_ERROR_IN_PROGRESS:
    r = EALREADY;
    break;
  case GNOME_VFS_ERROR_INTERRUPTED:
    r = EINTR;
    break;
  case GNOME_VFS_ERROR_FILE_EXISTS:
    r = EEXIST;
    break;
  case GNOME_VFS_ERROR_LOOP:
    r = ELOOP;
    break;
  case GNOME_VFS_ERROR_NOT_PERMITTED:
    r = EPERM;
    break;
  case GNOME_VFS_ERROR_IS_DIRECTORY:
    r = EISDIR;
    break;
  case GNOME_VFS_ERROR_NO_MEMORY:
    r = ENOMEM;
    break;
  case GNOME_VFS_ERROR_HOST_NOT_FOUND:
    r = EHOSTUNREACH;
    break;
  case GNOME_VFS_ERROR_INVALID_HOST_NAME:
    r = ENAMETOOLONG;
    break;
  case GNOME_VFS_ERROR_HOST_HAS_NO_ADDRESS:
    r = EHOSTUNREACH;
    break;
  case GNOME_VFS_ERROR_LOGIN_FAILED:
    r = EACCES;
    break;
  case GNOME_VFS_ERROR_CANCELLED:
    r = ECANCELED;
    break;
  case GNOME_VFS_ERROR_DIRECTORY_BUSY:
    r = EBUSY;
    break;
  case GNOME_VFS_ERROR_DIRECTORY_NOT_EMPTY:
    r = ENOTEMPTY;
    break;
  case GNOME_VFS_ERROR_TOO_MANY_LINKS:
    r = EMLINK;
    break;
  case GNOME_VFS_ERROR_READ_ONLY_FILE_SYSTEM:
    r = EROFS;
    break;
  case GNOME_VFS_ERROR_NOT_SAME_FILE_SYSTEM:
    r = EINVAL;
    break;
  case GNOME_VFS_ERROR_NAME_TOO_LONG:
    r = ENAMETOOLONG;
    break;
  case GNOME_VFS_ERROR_SERVICE_NOT_AVAILABLE:
    r = EADDRNOTAVAIL;
    break;
  case GNOME_VFS_ERROR_SERVICE_OBSOLETE:
    r = EINVAL;
    break;
  case GNOME_VFS_ERROR_PROTOCOL_ERROR:
    r = EPROTO;
    break;
  case GNOME_VFS_ERROR_NO_MASTER_BROWSER:
    r = ENOPROTOOPT;
    break;
  case GNOME_VFS_ERROR_NO_DEFAULT:
    r = EDOM;
    break;
  case GNOME_VFS_ERROR_NO_HANDLER:
    r = ENODATA;
    break;
  case GNOME_VFS_ERROR_PARSE:
    r = ENAMETOOLONG;
    break;
  case GNOME_VFS_ERROR_LAUNCH:
    r = ECHILD;
    break;
  case GNOME_VFS_ERROR_TIMEOUT:
    r = ETIMEDOUT;
    break;
  case GNOME_VFS_ERROR_NAMESERVER:
    r = EIDRM;
    break;
  case GNOME_VFS_ERROR_LOCKED:
    r = EDEADLK;
    break;
  case GNOME_VFS_ERROR_DEPRECATED_FUNCTION:
    r = ESTALE;
    break;
  case GNOME_VFS_NUM_ERRORS:
    r = ESTALE;
    break;
  default:
    r = ESTALE;
  }
  return -r;
}

static int xmp_getattr(const char *path, struct stat *stbuf)
{
    int rez;
    char *full_path;
    GnomeVFSResult result;
    GnomeVFSFileInfo *info;

    full_path = g_strdup_printf("%s/%s", mount_name, path);

    //g_debug("%s: %s", __FUNCTION__, full_path);

    info = gnome_vfs_file_info_new();
    result = gnome_vfs_get_file_info(full_path, info,
				     GNOME_VFS_FILE_INFO_DEFAULT | 
				     GNOME_VFS_FILE_INFO_GET_ACCESS_RIGHTS | 
				     GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
    if (result != GNOME_VFS_OK)
      {
	if (result == GNOME_VFS_ERROR_PROTOCOL_ERROR) {
	  /* work around for ssh returning this when file doesn't exists */
	  rez = -ENOENT;
	  goto dir_finished;
	}
	rez = gvfs2fuse(result);
	goto dir_finished;
      }
    
    memset(stbuf,0,sizeof(*stbuf));
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_SIZE)
      stbuf->st_size = info->size;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS)
      stbuf->st_mode = info->permissions;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_LINK_COUNT)
      stbuf->st_nlink = info->link_count;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS)
      stbuf->st_uid = info->uid;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_PERMISSIONS)
      stbuf->st_gid = info->gid;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_ATIME)
      stbuf->st_atime = info->atime;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME)
      stbuf->st_mtime = info->mtime;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_CTIME)
      stbuf->st_mtime = info->ctime;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_DEVICE)
      stbuf->st_dev = info->device;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_INODE)
      stbuf->st_ino = info->inode;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_IO_BLOCK_SIZE)
      stbuf->st_blksize = info->io_block_size;
    if (info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_BLOCK_COUNT)
      stbuf->st_blocks = info->block_count;

    stbuf->st_mode &= 0777;		/* keep just permissions */
    switch (info->type) {
    case GNOME_VFS_FILE_TYPE_REGULAR:
      stbuf->st_mode |= S_IFREG;
      break;
    case GNOME_VFS_FILE_TYPE_DIRECTORY:
      stbuf->st_mode |= S_IFDIR;
      break;
    case GNOME_VFS_FILE_TYPE_FIFO:
      stbuf->st_mode |= S_IFIFO;
      break;
    case GNOME_VFS_FILE_TYPE_SOCKET:
      stbuf->st_mode |= S_IFSOCK;
      break;
    case GNOME_VFS_FILE_TYPE_CHARACTER_DEVICE:
      stbuf->st_mode |= S_IFCHR;
      break;
    case GNOME_VFS_FILE_TYPE_BLOCK_DEVICE:
      stbuf->st_mode |= S_IFBLK;
      break;
    case GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK:
      stbuf->st_mode |= S_IFLNK;
      break;
    case GNOME_VFS_FILE_TYPE_UNKNOWN:
      break;
    }
    if (GNOME_VFS_FILE_INFO_SUID(info))
      stbuf->st_mode |= S_ISUID;
    if (GNOME_VFS_FILE_INFO_SGID(info))
      stbuf->st_mode |= S_ISGID;
    if (GNOME_VFS_FILE_INFO_STICKY(info))
      stbuf->st_mode |= S_ISVTX;
    //g_debug("mode: 0%o",stbuf->st_mode);

    rez = 0;

 dir_finished:
    gnome_vfs_file_info_unref(info);
    g_free(full_path);
    return rez;
}

static int xmp_readlink(const char *path, char *buf, size_t size)
{
    int res;
    char *full_path;
    GnomeVFSResult result;
    GnomeVFSFileInfo *info;

    full_path = g_strdup_printf("%s/%s", mount_name, path);
    //g_debug("%s %s", __FUNCTION__, full_path);

    info = gnome_vfs_file_info_new();
    result = gnome_vfs_get_file_info(full_path, info,
				     GNOME_VFS_FILE_INFO_DEFAULT);
    if (result != GNOME_VFS_OK)
      {
	res = gvfs2fuse(result);
	goto readlink_finished;
      }

    if (info->flags != GNOME_VFS_FILE_TYPE_SYMBOLIC_LINK) {
      res = -ENOLINK;
      goto readlink_finished;
    }

    g_snprintf(buf, size, "%s", info->symlink_name);
    res = 0;

 readlink_finished:
    g_free(full_path);
    gnome_vfs_file_info_unref(info);
    return res;
}

struct dir_trav_s {
  fuse_dirh_t dirh;
  fuse_dirfil_t dirfil;
};

gboolean dir_trav_fun (const gchar *rel_path,
		       GnomeVFSFileInfo *info,
		       gboolean recursing_will_loop,
		       gpointer data,
		       gboolean *recurse)
{
  struct dir_trav_s *dt = (struct dir_trav_s *) data;
  int ft;

  if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_TYPE) == 0) {
    ft = 0;
  }
  else {
    ft = info->type;
  }

  //g_debug("%s %s %d", __FUNCTION__, info->name, ft);

  dt->dirfil(dt->dirh, info->name, ft, info->inode);
  
  *recurse = FALSE;
  return TRUE;
}

static int xmp_getdir (const char * path, fuse_dirh_t dirh, fuse_dirfil_t dirfil)
{
  GnomeVFSResult result;
  GnomeVFSFileInfo *info;
  struct dir_trav_s dir_trav;
  int rez;
  char *full_path;

  //g_debug("%s", __FUNCTION__);

  full_path = g_strdup_printf("%s/%s", mount_name, path);

  info = gnome_vfs_file_info_new();
  result = gnome_vfs_get_file_info(full_path, info,
                                   GNOME_VFS_FILE_INFO_DEFAULT);
  if (result != GNOME_VFS_OK)
  {
    rez = gvfs2fuse(result);
    goto dir_finished;
  }

  if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_TYPE) == 0  ||
      (info->type & GNOME_VFS_FILE_TYPE_DIRECTORY) == 0) {
    rez = -ENOTDIR;
  }

  dir_trav.dirh = dirh;
  dir_trav.dirfil = dirfil;
  gnome_vfs_directory_visit(full_path, GNOME_VFS_FILE_INFO_DEFAULT,
			    GNOME_VFS_DIRECTORY_VISIT_DEFAULT |GNOME_VFS_DIRECTORY_VISIT_LOOPCHECK ,
			    dir_trav_fun, &dir_trav);
  rez = 0;

 dir_finished:
  g_free(full_path);
  gnome_vfs_file_info_unref(info);
  return rez;
}


static int xmp_mknod(const char *path, mode_t mode, dev_t rdev)
{
  int res;
  char *full_path;
  GnomeVFSResult result;
  GnomeVFSHandle *handle;
  
  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s %x", __FUNCTION__, full_path, mode);

#if 0
  /* check this since doesn't work as intended */
  if (mode & (S_IFCHR | S_IFBLK | S_IFIFO | S_IFSOCK)) {
    res = -EINVAL;
    goto mknod_finished;
  }
#endif

  result = gnome_vfs_create(&handle, full_path, GNOME_VFS_OPEN_NONE, TRUE, mode);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto mknod_finished;
  }
  res = 0;

 mknod_finished:
    g_free(full_path);
    return res;
}

static int xmp_mkdir(const char *path, mode_t mode)
{
  int res;
  char *full_path;
  GnomeVFSResult result;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);

  result = gnome_vfs_make_directory(full_path, mode);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto mkdir_finished;
  }
  res = 0;

 mkdir_finished:
    g_free(full_path);
    return res;

}

static int xmp_unlink(const char *path)
{
  int res;
  char *full_path;
  GnomeVFSResult result;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);


  result = gnome_vfs_unlink(full_path);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto unlink_finished;
  }
  res=0;

 unlink_finished:
    g_free(full_path);
    return res;
}

static int xmp_rmdir(const char *path)
{
  int res;
  char *full_path;
  GnomeVFSResult result;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);

  result = gnome_vfs_remove_directory(full_path);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto rmdir_finished;
  }
  res=0;

 rmdir_finished:
    g_free(full_path);
    return res;
}

static int xmp_symlink(const char *from, const char *to)
{
  int res;
  char *full_path_from, *full_path_to;
  GnomeVFSResult result;
  GnomeVFSURI* uri_to;

  full_path_from = g_strdup_printf("%s/%s", mount_name, from);
  full_path_to = g_strdup_printf("%s/%s", mount_name, to);
  //g_debug("%s %s %s", __FUNCTION__, full_path_from, full_path_to);

  if ((uri_to = gnome_vfs_uri_new(full_path_to)) == NULL) {
    res = -ENOENT;
    goto symlink_finished;
  }

  result = gnome_vfs_create_symbolic_link(uri_to, full_path_from);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto symlink_finished;
  }
  res=0;

 symlink_finished:
  gnome_vfs_uri_unref(uri_to);
  g_free(full_path_from);
  g_free(full_path_to);
  return res;
}

static int xmp_rename(const char *from, const char *to)
{
  int res;
  char *full_path_from, *full_path_to;
  GnomeVFSResult result;

  full_path_from = g_strdup_printf("%s/%s", mount_name, from);
  full_path_to = g_strdup_printf("%s/%s", mount_name, to);
  //g_debug("%s %s %s", __FUNCTION__, full_path_from, full_path_to);

  result = gnome_vfs_move(full_path_from, full_path_to, TRUE);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto rename_finished;
  }
  res=0;

 rename_finished:
    g_free(full_path_from);
    g_free(full_path_to);
    return res;
}

#if 0
static int xmp_link(const char *from, const char *to)
{
    return -ENOSYS;
}
#endif

static int xmp_chmod(const char *path, mode_t mode)
{
  int res;
  char *full_path;
  GnomeVFSResult result;
  GnomeVFSFileInfo *info;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);
  info = gnome_vfs_file_info_new();

  info->permissions = mode;
  result = gnome_vfs_set_file_info(full_path, info, GNOME_VFS_SET_FILE_INFO_PERMISSIONS);
  if (result != GNOME_VFS_OK)
    {
      res = gvfs2fuse(result);
      goto chmod_finished;
    }
  res = 0;

 chmod_finished:
    gnome_vfs_file_info_unref(info);
    g_free(full_path);
    return res;
}

static int xmp_chown(const char *path, uid_t uid, gid_t gid)
{
  int res;
  char *full_path;
  GnomeVFSResult result;
  GnomeVFSFileInfo *info;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);
  info = gnome_vfs_file_info_new();

  info->uid = uid;
  info->gid = gid;
  result = gnome_vfs_set_file_info(full_path, info, GNOME_VFS_SET_FILE_INFO_OWNER);
  if (result != GNOME_VFS_OK)
    {
      res = gvfs2fuse(result);
      goto chown_finished;
    }
  res = 0;

 chown_finished:
    gnome_vfs_file_info_unref(info);
    g_free(full_path);
    return res;
}

static int xmp_truncate(const char *path, off_t size)
{
  int res;
  char *full_path;
  GnomeVFSResult result;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);

  result = gnome_vfs_truncate(full_path, size);
  if (result != GNOME_VFS_OK)
  {
    res = gvfs2fuse(result);
    goto truncate_finished;
  }
  res=0;

 truncate_finished:
    g_free(full_path);
    return res;
}

static int xmp_utime(const char *path, struct utimbuf *buf)
{
  int res;
  char *full_path;
  GnomeVFSResult result;
  GnomeVFSFileInfo *info;

  full_path = g_strdup_printf("%s/%s", mount_name, path);
  //g_debug("%s %s", __FUNCTION__, full_path);

  info = gnome_vfs_file_info_new();
  result = gnome_vfs_get_file_info(full_path, info,
				   GNOME_VFS_FILE_INFO_DEFAULT);
  if (result != GNOME_VFS_OK)
    {
      res = gvfs2fuse(result);
      goto utime_finished;
    }
  info->atime = buf->actime;
  info->mtime = buf->modtime;
  result = gnome_vfs_set_file_info(full_path, info, GNOME_VFS_SET_FILE_INFO_TIME);
  if (result != GNOME_VFS_OK)
    {
      res = gvfs2fuse(result);
      goto utime_finished;
    }
  res = 0;

 utime_finished:
    gnome_vfs_file_info_unref(info);
    g_free(full_path);
    return res;
}


static int xmp_open(const char *path, struct fuse_file_info *fi)
{
    int res;
    char *full_path;
    GnomeVFSOpenMode flags = GNOME_VFS_OPEN_RANDOM;
    GnomeVFSResult result;
    GnomeVFSHandle *handle = (GnomeVFSHandle *) fi->fh;

    full_path = g_strdup_printf("%s/%s", mount_name, path);

    //g_debug("%s: %s", __FUNCTION__, full_path);

    if (fi->flags & (O_RDONLY | O_RDWR))
      flags |= GNOME_VFS_OPEN_READ;
    if (fi->flags & (O_WRONLY | O_RDWR))
      flags |= GNOME_VFS_OPEN_WRITE;
    
    result = gnome_vfs_open(&handle, full_path, flags);
    if (result != GNOME_VFS_OK) {
      res = gvfs2fuse(result);
      goto open_finished;
    }

    fi->fh = (unsigned long) handle;
    res = 0;

 open_finished:
    g_free(full_path);
    return res;
}

static int xmp_read(const char *path, char *buf, size_t size, off_t offset,
                    struct fuse_file_info *fi)
{
    int res;
    GnomeVFSResult result;
    GnomeVFSHandle *handle = (GnomeVFSHandle *) fi->fh;
    GnomeVFSFileSize nread;
    (void) path;

    //g_debug("%s %s", __FUNCTION__, path);

    result = gnome_vfs_seek(handle, GNOME_VFS_SEEK_START, offset);
    if (result != GNOME_VFS_OK) {
      res = gvfs2fuse(result);
      return res;
    }
    result = gnome_vfs_read(handle, buf, size, &nread);
    if (result != GNOME_VFS_OK) {
      res = gvfs2fuse(result);
      return res;
    }

    return nread;
}

static int xmp_write(const char *path, const char *buf, size_t size,
                     off_t offset, struct fuse_file_info *fi)
{
    int res;
    GnomeVFSResult result;
    GnomeVFSHandle *handle = (GnomeVFSHandle *) fi->fh;
    GnomeVFSFileSize nwrite;
    (void) path;

    //g_debug("%s %s", __FUNCTION__, path);

    result = gnome_vfs_seek(handle, GNOME_VFS_SEEK_START, offset);
    if (result != GNOME_VFS_OK) {
      res = gvfs2fuse(result);
      return res;
    }
    result = gnome_vfs_write(handle, buf, size, &nwrite);
    if (result != GNOME_VFS_OK) {
      res = gvfs2fuse(result);
      return res;
    }

    return nwrite;
}

#if 0
static int xmp_statfs(const char *path, struct statfs *stbuf)
{
  //g_debug("%s", __FUNCTION__);
  return -ENOSYS;
}
#endif

static int xmp_release(const char *path, struct fuse_file_info *fi)
{
    GnomeVFSHandle *handle = (GnomeVFSHandle *) fi->fh;
    (void) path;

    //g_debug("%s %s", __FUNCTION__, path);
    gnome_vfs_close(handle);

    return 0;
}

#if 0
static int xmp_fsync(const char *path, int isdatasync,
                     struct fuse_file_info *fi)
{
  return -ENOSYS;
}
#endif

static struct fuse_operations xmp_oper = {
    .getattr	= xmp_getattr,
    .readlink	= xmp_readlink,
    .getdir     = xmp_getdir,
    .mknod	= xmp_mknod,
    .mkdir	= xmp_mkdir,
    .symlink	= xmp_symlink,
    .link	= xmp_symlink,
    .unlink	= xmp_unlink,
    .rmdir	= xmp_rmdir,
    .rename	= xmp_rename,
    .chmod	= xmp_chmod,
    .chown	= xmp_chown,
    .truncate	= xmp_truncate,
    .utime	= xmp_utime,
    .open	= xmp_open,
    .read	= xmp_read,
    .write	= xmp_write,
    //.statfs	= xmp_statfs,
    .release	= xmp_release,
    //.fsync	= xmp_fsync,
};

int main(int argc, char *argv[])
{
  int fuse_ret, i, j, inou;
  GnomeVFSResult result;
  GnomeVFSFileInfo *info;

  /* initialize GnomeVFS */
  if (!gnome_vfs_init()) {
    g_error("could not initialize GnomeVFS\n");
    return 1;
  }

  inou = 0;
  for (i=1; i<argc; i++) {
    if (argv[i][0] != '-') {
      if (inou == 0) {
	mount_name = argv[i];
	argc--;
	for(j=i; j<argc; j++) 
	  argv[j]=argv[j+1];
	i--;
      }
      inou++;
    }
  }
  if (inou != 2 && inou != 0) {
    g_printerr("Syntax: %s [options] [gnome vfs uri] [mount point]\n", argv[0]);
    goto gnome_shutdown;
  }

  /* test if we are actually mounting a directory, otherwise it has little sense */
  info = gnome_vfs_file_info_new();
  result = gnome_vfs_get_file_info(mount_name, info,
                                   GNOME_VFS_FILE_INFO_DEFAULT);
  if (result != GNOME_VFS_OK)
  {
    fuse_ret=1;
    g_printerr("%s: cannot open URI (Gnome VFS reports: %s)\n", 
	       argv[0], gnome_vfs_result_to_string(result));
    goto gnome_shutdown;
  }

  if ((info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_TYPE) == 0  ||
      (info->type & GNOME_VFS_FILE_TYPE_DIRECTORY) == 0) {
    fuse_ret=2;
    g_printerr("%s: the URI is not a directory\n", argv[0]);
    goto gnome_shutdown;
  }

  /* now do thee real mount */
  fuse_ret = fuse_main(argc, argv, &xmp_oper);

  /* shut down GnomeVFS */
 gnome_shutdown:
  gnome_vfs_shutdown();

  return fuse_ret;
}
